/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2017, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

    http://www.johncon.com/ntropix/
    http://www.johncon.com/

------------------------------------------------------------------------------

spnmeaUSB0.c is for asynchronous raw serial port communication. The
default port is /dev/ttyUSB0 at 4800 baud. The default semaphore lock
file is /var/lock/LCK..ttyUSB0, which has an fcntl(2) advisory lock
applied, also. The serial port is controlled through termios(3), and
is restored to original values on exit. The serial port child read
process is created by fork(2) from the parent write process, and both
processes respond to SIGINT = ^C and SIGTERM signals. Asynchronous
serial port control is implemented with select(2). The input file,
(usually stdin,) uses canonical input for flow control.

This program is a rework of sprawS0.c in sp.tar.gz, with the default
device changed to /dev/ttyUSB0, and the baud rate set to 4800 Baud, as
per the NEMA standard. The location coordinates and UMT are a record
of the form: "$GPGLL,3716.6051,N,12158.5555,W,182214.000,A,A*49",
(without the double quotes.) Note the "^$GPGLL," keyword.

The default values are for Linux 7.X, Wheezy, and an ND-100S GPS USB
dongle with an SiRF StarIII GPS chipset, and, ID 067b:2303 Prolific
Technology, Inc. PL2303 Serial Port:

    idVendor  0x067b Prolific Technology, Inc.
    idProduct 0x2303 PL2303 Serial Port

$Revision: 0.0 $
$Date: 2017/06/18 19:36:00 $
$Id: spnmeaUSB0.c,v 0.0 2017/06/18 19:36:00 john Exp $
$Log: spnmeaUSB0.c,v $
Revision 0.0  2017/06/18 19:36:00  john
Initial revision


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/wait.h>

#define _POSIX_SOURCE 1 /* POSIX compliant source */

#define BAUDRATE B4800 /* baud rate for serial port communication */
#define SPDEVICE "/dev/ttyUSB0" /* default serial port device */
#define LOCKFILE "/var/lock/LCK..ttyUSB0" /* default serial port lock file name */

#define BUFLEN BUFSIZ /* size of record buffers */

static char rcsid[] = "$Id: spnmeaUSB0.c,v 0.0 2017/06/18 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2017, John Conover, All Rights Reserved"; /* the copyright banner */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Serial port ttyUSB0 communication\n",
    "Usage: spnmeaUSB0 [-d device] [-i infilename] [-l lockfile] [-o outfilename] [-v]\n",
    "    -d device, device name, (/dev/ttyUSB0)\n",
    "    -i infilename, input filename, (stdin)\n",
    "    -l lockfile, lock file name, (/var/lock/LCK..ttyUSB0)\n",
    "    -o outfilename, output file name, (stdout)\n",
    "    -v, print the program's version information\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening serial device\n",
    "Error getting serial device attributes\n",
    "Error installing signal handler\n",
    "Error flushing serial device attributes\n",
    "Error setting serial device attributes\n",
    "Error writing to serial device\n",
    "Error resetting serial device attributes\n",
    "Error closing serial device\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error creating lock file\n",
    "Error removing lock file\n",
    "Error forking to child read process\n",
    "Error reading from serial device\n",
    "Error locking lock file\n",
    "Error writing to lock file\n",
    "Error opening output file\n",
    "Error closing output file\n",
    "Error reading input file\n",
    "Error writing to output file\n",
    "Error enabling blocking on serial device\n",
    "Error flushing output file\n",
    "Error setting baud rate\n",
    "Error flushing serial device\n",
    "Error write buffer overflow\n",
    "Error select\n"
};

#define NOERROR 0 /* negative error values, one for each index in the error message array */
#define EARGS -1
#define EOPEN -2
#define EGATTR -3
#define ESIG -4
#define EFLUSH -5
#define ESATTR -6
#define EWRITE -7
#define ERATTR -8
#define ECLOSE -9
#define EIOPEN -10
#define EICLOSE -11
#define ELOCKF -12
#define ELOCKR -13
#define EFORK -14
#define EREAD -15
#define EFCNL -16
#define EWLCK -17
#define EOOPEN -18
#define EOCLOSE -19
#define ERDIN -20
#define EOWRITE -21
#define EBLOCK -22
#define EOFLUSH -23
#define EBAUD -24
#define EDFLUSH -25
#define EWROVFL -26
#define ESEL -27

#ifdef __STDC__

static void print_message (int pretval); /* print any error messages */
static void spclose (int err); /* interrupt handler */
static int child_read (int fdd); /* child read loop, process all read records from the modem */
static int parent_write (int fdd, int infiled); /* parent write loop, process all write records to the modem */

#else

static void print_message (); /* print any error messages */
static void spclose (); /* interrupt handler */
static int child_read (); /* child read loop, process all read records from the modem */
static int parent_write (); /* parent write loop, process all write records to the modem */

#endif

static char serial_lockname[] = {LOCKFILE}, /* default lock file name */
            *lockname = serial_lockname, /* default lock file name */
            serial_device[] = {SPDEVICE}; /* default serial port device name */

static int retval = NOERROR, /* return value, assume no error */
           fd = -1, /* serial port device file descriptor */
           infile = STDIN_FILENO, /* input file descriptor */
           outfile = STDOUT_FILENO, /* output file file descriptor */
           semaphorelock = -1, /* lock file file descriptor */
           state = 0, /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
           remove_lockfile = 0; /* remove lock file, 0 = no, 1 = yes */

static pid_t pid_write = (pid_t) 0, /* pid of parent write process */
             pid_read = (pid_t) 0; /* pid of child read process */

static struct termios oldtio; /* original device attributes, restored on exit */

static struct flock lockfile; /* lock file structure */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    static char pid_str[BUFLEN]; /* pid string */

    char *device = serial_device, /* default serial port device name */
         *in_name = (char *) 0, /* input file name */
         *out_name = (char *) 0; /* output file name */

    int c; /* command line switch */

    struct termios newtio; /* device attributes used by this program */

    pid_write = getpid (); /* save the pid of parent write process */

    while ((c = getopt (argc, argv, "d:i:l:o:v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'd': /* request for device name? */

                device = optarg; /* yes, save the device name */
                break;

            case 'i': /* request for input file name? */

                in_name = optarg; /* yes, save the input file name */
                break;

            case 'l': /* request for lock file name? */

                lockname = optarg; /* yes, save the lock file name */
                break;

            case 'o': /* request for output file name? */

                out_name = optarg; /* yes, save the output file name */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = ESIG; /* assume error installing signal handler */

        if (signal (SIGINT, spclose) != SIG_ERR) /* install the interrupt handler for SIGINT = ^C */
        {

            if (signal (SIGTERM, spclose) != SIG_ERR) /* install the interrupt handler for SIGTERM */
            {

                if (signal (SIGCHLD, spclose) != SIG_ERR) /* install the interrupt handler for SIGCHLD */
                {
                    lockfile.l_type = F_WRLCK; /* exclusive write lock the entire file */
                    lockfile.l_whence = (short int) SEEK_SET;
                    lockfile.l_start = (off_t) 0;
                    lockfile.l_len = (off_t) 0;

                    retval = ELOCKF; /* assume error creating lock file */

                    if ((semaphorelock = open (lockname, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the lock file exclusively */
                    {
                        remove_lockfile = 1; /* set the remove lock file, 0 = no, 1 = yes */
                        retval = NOERROR; /* assume no error */
                    }

                    else if ((semaphorelock = open (lockname, O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the lock file if it exists */
                    {
                        retval = NOERROR; /* assume no error */
                    }

                    if (retval == NOERROR) /* any errors? */
                    {
                        retval = EFCNL; /* assume error locking lock file */

                        if (fcntl (semaphorelock, F_SETLK, &lockfile) == 0) /* exclusive write lock the lock file */
                        {
                            remove_lockfile = 1; /* set the remove lock file, 0 = no, 1 = yes */
                            retval = EWLCK; /* assume error writing to lock file */

                            if (ftruncate (semaphorelock, 0) == 0) /* truncate the lock file */
                            {
                                (void) sprintf (pid_str, "%10d\n", pid_write); /* the pid string is the PID right justified to 10 characters plus a '\n' EOL */

                                if (write (semaphorelock, pid_str, strlen (pid_str)) >= 0) /* write the serial port dialer string to the serial port */
                                {
                                    retval = EOPEN; /* assume error opening device */

                                    if ((fd = open (device, O_RDWR | O_NOCTTY | O_NDELAY)) >= 0) /* open the serial port device file descriptor */
                                    {
                                        retval = EBLOCK; /* assume error enabling blocking on serial device */

                                        if (fcntl (fd, F_SETFL, O_NONBLOCK) == 0) /* enable read blocking, i.e., read () waits until a character is in the buffer */
                                        {
                                            retval = EFLUSH; /* assume error flushing device attributes */

                                            if (tcflush (fd, TCIOFLUSH) == 0) /* flush device input/output */
                                            {
                                                retval = EGATTR; /* assume error getting device attributes */

                                                if (tcgetattr (fd, &oldtio) == 0) /* save original port settings */
                                                {
                                                    state = 1; /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
                                                    memcpy (&newtio, &oldtio, sizeof (newtio)); /* copy the original device attributes */
                                                    retval = EBAUD; /* assume error setting baud rate */

                                                    if (cfsetispeed (&newtio, BAUDRATE) == 0) /* set the baud rate */
                                                    {

                                                        if (cfsetospeed (&newtio, BAUDRATE) == 0)
                                                        {

                                                            /* raw terminal set up, see "man cfmakeraw(3)" for particulars */

                                                            newtio.c_cflag |= (CLOCAL | CREAD); /* local mode, enable receiver */
                                                            newtio.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
                                                            newtio.c_oflag &= ~OPOST;
                                                            newtio.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
                                                            newtio.c_cflag &= ~(CSIZE|PARENB); /* No parity (8N1) */
                                                            newtio.c_cflag |= CS8;
                                                            newtio.c_cc[VTIME] = 0;
                                                            newtio.c_cc[VMIN] = 1;

                                                            retval = ESATTR; /* assume error setting device attributes */

                                                            if (tcsetattr (fd, TCSANOW, &newtio) == 0) /* set the device attributes */
                                                            {
                                                                retval = EFLUSH; /* assume error flushing device attributes */

                                                                if (tcflush (fd, TCIOFLUSH) == 0) /* flush device input/output */
                                                                {
                                                                    retval = EOOPEN; /* assume error opening output file */

                                                                    if ((outfile = (out_name == (char *) 0) ? outfile : open (out_name, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH )) >= 0) /* open the output file */
                                                                    {
                                                                        retval = EIOPEN; /* assume error opening file */

                                                                        if ((infile = (in_name == (char *) 0) ? infile : open (in_name, O_RDWR | O_NOCTTY)) >= 0) /* open the input file */
                                                                        {
                                                                            retval = NOERROR; /* assume no error */

                                                                            if ((pid_read = fork ()) < 0) /* fork the child read process */
                                                                            {
                                                                                retval = EFORK; /* assume error forking to child read process */
                                                                            }

                                                                            else if (pid_read == 0) /* child or parent process, (pid_read is zero in the child process, the child process' PID in the parent)? */
                                                                            {

                                                                                /* start of child read process; pid_write still contains the PID of the parent write process */

                                                                                retval = child_read (fd); /* child read loop, process all read records from the modem */

                                                                                /* end of the child read process, fall through and exit */
                                                                            }

                                                                            else /* wasn't child process, must be parent */
                                                                            {

                                                                                /* start of the parent write process; pid_read contains the PID of the child read process */

                                                                                retval = parent_write (fd, infile); /* parent write loop, process all write records to the modem */

                                                                                /* end of the parent write process, fall through and exit */

                                                                            }

                                                                        }

                                                                    }

                                                                }

                                                            }

                                                        }

                                                    }

                                                }

                                            }

                                        }

                                    }

                                }

                            }

                        }

                    }

                }

            }

        }

    }

    spclose (0); /* shutdown */
    exit (-retval); /* exit with the error value, formality: spclose () did the exit */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int pretval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer pretval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer pretval:

        1) If pretval is zero, print nothing-normal/successful program
           exit.

        2) Else if pretval is unity, print help.

        3) Else pretval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int pretval)

#else

static void print_message (pretval)
int pretval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (pretval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[pretval]);
            break;

    }

}

/*

Interrupt handler and shutdown.

static void spclose (int err);

I) Prerequisites:

    A) All file opening and serial port initialization is done by the
       parent write process, (excepting the log file, which opened and
       closed dynamically by the child read process-and is normally
       closed,)

        1) The parent write process should close all files, (except
           the log file,) and restore the serial port.

        2) The child read process should close the log file.

    B) Called with SIGINT = ^C, or, SIGTERM, or, SIGCHLD interrupts,
       or, with a negative argument representing an internal error,
       or, an argument of zero for no errors or interrupts.

    C) There are "edge" issues, (i.e., a flag is tested to see if
       something needs shutting down, and the something is shut down,
       but the program is interrupted-perhaps by SIGINT = ^C
       bashing-before the flag gets reset,) There are no error message
       printed in the shut down section of this routine-everything is
       shut down, and at the end of this routine, the flags are
       queried for a successful shutdown, and shutdown errors, if any,
       printed. The idea being that if an "edge" issue occurred, the
       shut down would occur twice, (presumably giving an error the
       second time, which would not be printed, but as long as the
       flag is reset, no error would be printed.)

    D) The kill of the child read by the parent write process, (and
       vice verse,) is complex:

        1) static passed = reset
        2) if parent
            a) if err is SIGCHLD
                i) wait for the child read process to die
            b) if passed = reset
                i) reset all things in the parent process
                ii) print any errors
                iii) set passed
                iv) if err not SIGCHLD and child read process active,
                    kill child read process
        3) else if child
            a) if passed = reset
                i) close the log file
                ii) print any errors
                iii) set passed
                iv) kill parent
        4) restore all signals

II) Data structures:

    A) if the pid is equal to pid_write then the parent write process
       is executing this routine, else, it is the child read process.

    B) The parent write and child read process have a separate copy of
       the int passed flag-which means a complete pass has been made
       through this routine for the client or parent.

III) Function execution:

    A) Upon entry:

        1) If the PID of the process is the parent write process:

            a) Shut down all devices opened by the parent write
               process.

            b) Print any pending errors that occurred before the
               program entered this routine.

            c) Print any errors after the shutdown is complete.

            d) If the call to this routine was not caused by a dead
               child read process, and the child process is actve,
               kill the child read process.

        2) If the PID of the process is the child read process:

            a) Shut down all devices opened by the child read process.

            b) Print any pending errors that occurred before the
               program entered this routine.

            c) Print any errors after the shutdown is complete.

            d) Kill the parent write process.

        3) Restore the signal handlers for SIGINT = ^C and SIGTERM.

There is no return from this routine. The routine exits with a value
of zero if no errors or interrupts occurred, or, the value of the
interrupt, (a positive number,) if an interrupt occurred, or, with a
negative error number if an an error occurred.

*/

#ifdef __STDC__

static void spclose (int err)

#else

static void spclose (err)
int err;

#endif

{
    static int passed = 0; /* passed this routine, 0 = no, 1 = yes, used by both parent write and child read processes, but are different copies */

    int exit_value,
        local_retval = NOERROR; /* assume no error shutting things down */

    if (pid_write == getpid ()) /* parent write process? */
    {

        if (err == SIGCHLD) /* child read process died? */
        {
            (void) wait ((int *) 0); /* free the child read process */
        }

        if (passed == 0) /* passed this routine, 0 = no, 1 = yes? */
        {

            if (state == 1) /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
            {
                (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */

                if (tcsetattr (fd, TCSANOW, &oldtio) == 0) /* reset the device attributes */
                {
                    state = 0; /* state of serial port setup, 0 = not set up, 1 = original device attributes successfully saved */
                    (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */
                }

            }

            if (fd > 0) /* serial port device file descriptor successfully opened? */
            {
                (void) tcflush (fd, TCIOFLUSH); /* flush device input/output */

                if (close (fd) == 0) /* yes, close the device */
                {
                    fd = -1; /* reset the serial port device file descriptor */
                }

            }

            if (infile != STDIN_FILENO) /* input file open? */
            {

                if (close (infile) == 0) /* close the input file */
                {
                    infile = STDIN_FILENO; /* reset the input file descriptor */
                }

            }

            if (outfile != STDOUT_FILENO) /* output file open? */
            {

                if (close (outfile) == 0) /* close the output file */
                {
                    outfile = STDOUT_FILENO; /* reset the output file descriptor */
                }

            }

            if (semaphorelock >= 0) /* lock file open? */
            {
                lockfile.l_type = F_UNLCK; /* unlock the file */
                (void) fcntl (semaphorelock, F_SETLKW, &lockfile); /* unlock the lock file */

                if (close (semaphorelock) == 0) /* close the lock file */
                {
                    semaphorelock = -1; /* reset the lock file file descriptor */

                    if (remove_lockfile == 1) /* remove lock file, 0 = no, 1 = yes, set? */
                    {

                        if (unlink (lockname) == 0) /* remove the lock file */
                        {
                            remove_lockfile = 0; /* remove lock file, 0 = no, 1 = yes */
                        }

                    }

                }

            }

            print_message (-retval); /* print any error messages from the program, (there can only be zero or one) */

            /* print any errors shutting things down */

            if (state != 0) /* state of serial port still setup? */
            {print_message (-ERATTR); local_retval = ERATTR;} /* print the error message, set the local error shutting things down */

            if (semaphorelock != -1) /* lock file open? */
            {print_message (-ELOCKR); local_retval = ELOCKR;} /* print the error message, set the local error shutting things down */

            /* probably don't care about these-exit () will handle them */

            if (fd != -1) /* serial port device file descriptor still open? */
            {print_message (-ECLOSE); local_retval = ECLOSE;} /* print the error message, set the local error shutting things down */

            if (infile != STDIN_FILENO) /* input file open? */
            {print_message (-EICLOSE); local_retval = EICLOSE;} /* print the error message, set the local error shutting things down */

            if (outfile != STDOUT_FILENO) /* output file open? */
            {print_message (-EOCLOSE); local_retval = EOCLOSE;} /* print the error message, set the local error shutting things down */

            passed = 1; /* passed this routine, 0 = no, 1 = yes, used by both parent write and child read processes, but are different copies */

            if (err != SIGCHLD) /* child read process died? */
            {

                if (pid_read != (pid_t) 0) /* child read process forked? */
                {
                    (void) kill (pid_read, SIGTERM); /* kill the child read process */
                }

            }

        }

    }

    else /* no, child read process */
    {

        if (passed == 0) /* passed this routine, 0 = no, 1 = yes? */
        {
            print_message (-retval); /* print any error messages from the program, (there can only be zero or one) */
            passed = 1; /* passed this routine, 0 = no, 1 = yes, used by both parent write and child read processes, but are different copies */

            if (pid_write != (pid_t) 0) /* parent write process active? */
            {
                (void) kill (pid_write, SIGTERM); /* kill the parent write process */
            }

        }

    }

    /* probably don't care about these-exit () will handle them */

    (void) signal (SIGCHLD, SIG_DFL); /* restore default interrupt for SIGCHLD */
    (void) signal (SIGTERM, SIG_DFL); /* restore default interrupt for SIGTERM */
    (void) signal (SIGINT, SIG_DFL); /* restore default for SIGINT = ^C interrupt */

    exit_value = err; /* assume routine called externally, i.e., SIGINT == ^C or SIGTERM */

    if (exit_value == 0) /* no? */
    {
        exit_value = retval; /* assume error in main body of program, before this routine */

        if (retval == 0) /* no? */
        {
            exit_value = local_retval; /* assume error shutting things down */
        }

    }

    exit (exit_value); /* exit with the value of the interrupt, the value of an error in the main body of this program, the value of an error in this routine, or zero for no error */
}

/*

Parent write loop, process all records from the input file, (usually,
stdin,) to the serial device.

static int parent_write (int fdd, int infiled);

I) Data structure:

    A) The serial port device file descriptor, fdd, should be open, and
       available.

    B) The input file descriptor, infiled, should be open and
       available; it is probably stdin, (but can be a named pipe,
       etc.)

II) Function execution:

    A) This routine is a "spin loop" that waits for a signal, or, an
       error to terminate the program, while reading characters from
       the input file descriptor, infiled.

    B) Input characters from the input file are printed to the output
       file.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int parent_write (int fdd, int infiled)

#else

static int parent_write (fdd, infiled)
int fdd;
int infiled;

#endif

{
    static char buffer[BUFLEN + 1]; /* input write buffer */

    int local_retval = NOERROR; /* local return value, (so ERDIN = "Error reading input file" will not go to spclose () on a SIGINT = ^C or SIGTERM "edge" in this routine,) assume no error */

    ssize_t cnt; /* count of characters read */

    buffer[BUFLEN] = '\0'; /* write buffer sentinel */

    while (local_retval == NOERROR) /* while no error, spin though getting charcters from the serial device, (actually, waiting for a SIGINT = ^C or SIGTERM) */
    {
        local_retval = ERDIN; /* assume error reading input file */

        if ((cnt = read (infiled, buffer, BUFLEN)) > 0) /* read characters from the input file */
        {
            local_retval = EWROVFL; /* assume error write buffer overflow */

            if (strchr (buffer, '\n') != (char *) 0) /* buffer have an EOL character before the sentinel? */
            {
                local_retval = EWRITE; /* assume error writing to serial device */

                if (write (fdd, buffer, (size_t) cnt) == cnt) /* write the characters to the serial device */
                {
                    local_retval = EDFLUSH; /* assume error flushing serial device */

                    if (tcdrain (fdd) == 0) /* flush the serial device */
                    {
                        local_retval = NOERROR; /* assume no error */
                    }

                }

            }

        }

    }

    return (retval = local_retval); /* return any errors */
}

/*

Child read loop, process all records from the modem.

static int child_read (int fdd);

I) Data structure:

    A) The serial port device file descriptor, fdd, should be open, and
       available.

II) Function execution:

    A) This routine is a "spin loop" that waits for a a character from
       the serial device using select(2), a signal, or, an error to
       terminate the program, while reading characters from the serial
       device file descriptor, fdd.

    B) Input characters from the serial device are written to the
       output file.

Returns any errors encountered, NOERROR on success.

*/

#ifdef __STDC__

static int child_read (int fdd)

#else

static int child_read (fdd)
int fdd;

#endif

{
    static char buffer[BUFLEN]; /* input read buffer */

    int local_retval = NOERROR; /* local return value, (so EREAD = "Error reading from device" will not go to spclose () on a SIGINT = ^C or SIGTERM "edge" in this routine,) assume no error */

    ssize_t cnt; /* count of characters read */

    fd_set fds; /* select () struct */

    FD_ZERO (&fds); /* null the select () struct */
    FD_SET (fdd, &fds); /* and install it for the serial file descriptor */

    while (local_retval == NOERROR) /* while no error, spin though getting charcters from the serial device, (actually, waiting for a SIGINT = ^C or SIGTERM) */
    {
        local_retval = ESEL; /* assume select error */

        if (select (fdd + 1, &fds, 0, 0, 0) != -1) /* wait for characters from the serial device */
        {
            local_retval = NOERROR; /* assume no error */

            if (FD_ISSET (fd, &fds)) /* data ready? */
            {
                local_retval = EREAD; /* assume error reading from device */

                if ((cnt = read (fdd, buffer, BUFLEN)) > 0) /* read characters from the serial device */
                {
                    local_retval = EOWRITE; /* assume error writing to output file */

                    if (write (outfile, buffer, (size_t) cnt) > 0) /* write the characters to the output device */
                    {

                        if (outfile == STDOUT_FILENO) /* output file stdout, (can't flush stdout)? */
                        {
                            local_retval = NOERROR; /* assume no error */
                        }

                        else
                        {

                            if (fsync (outfile) == 0) /* flush the output file */
                            {
                                local_retval = NOERROR; /* assume no error */
                            }

                        }

                    }

                }

            }

        }

    }

    return (retval = local_retval); /* return any errors */
}
